from face2face import Face2Face
from media_toolkit import VideoFile
import cv2

f2f = Face2Face()

# test single face swap
#source_img = cv2.imread("test_imgs/test_face_1.jpg")
#target_img = cv2.imread("test_imgs/test_face_2.jpg")
#
#swapped = f2f.swap_one_image(source_img, target_img)
#cv2.imshow("swapped_face", swapped)
#
## test embedding face swap
#f2f.add_reference_face("hagrid", source_img, save=True)
#swapped = f2f.swap_from_reference_face("hagrid", target_img)

# test video face swap
#source_img = cv2.imread("test_imgs/test_face_3.jpg")
#f2f.add_reference_face("black_woman", source_img, save=True)
vf = VideoFile().from_file("test_imgs/test_vid_1.mp4")
swapped = f2f.swap_video(face_name="hagrid", target_video=vf)
swapped.save("test_imgs/swapped_vid_1.mp4")

# debug point
a = 1



import os

import cv2
from tqdm import tqdm

from face2face.settings import OUTPUT_DIR, MODELS_DIR
from face2face.utils import get_files_in_dir


def make_video_from_images(
    image_paths: list[str], outpath: str = None, frame_rate: int = 60
) -> None:
    """creates a video from a list of images"""

    if outpath is None:
        outpath = os.path.join(OUTPUT_DIR, "output.mp4")

    # get image dimensions
    firstimg = cv2.imread(image_paths[0])
    height, width, layers = firstimg.shape

    # create video writer
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    video = cv2.VideoWriter(outpath, fourcc, frame_rate, (width, height))

    # write images to video
    for image_path in tqdm(image_paths, desc="writing images to video"):
        image = cv2.imread(image_path)
        video.write(image)

    # close video writer
    video.release()


def make_video_from_image_folder(
    image_folder: str, outpath: str = None, frame_rate: int = 60
) -> None:
    """creates a video from a folder of images"""
    image_paths = get_files_in_dir(image_folder, [".jpeg", ".jpg", ".png"])
    # order images by creation date otherwise the video will be out of order
    image_paths.sort(key=os.path.getmtime)
    make_video_from_images(image_paths, outpath, frame_rate)


def video2images(video_path: str, outpath: str = None) -> None:
    """creates a video from a list of images
    video_path: path to video file in mp4
    outpath: folder to save the images (no trailing /)
    """

    if outpath is None:
        outpath = OUTPUT_DIR

    if not os.path.isdir(outpath):
        os.makedirs(outpath)

    vidcap = cv2.VideoCapture(video_path)
    success, image = vidcap.read()
    count = 0
    while success:
        cv2.imwrite(outpath + "/frame%d.jpg" % count, image)  # save frame as JPEG file
        success, image = vidcap.read()
        print('Read a new frame: ', success)
        count += 1

    # write framerate to file
    framerate = vidcap.get(cv2.CAP_PROP_FPS)
    with open(os.path.join(outpath, "_framerate.txt"), "w") as f:
        f.write(str(framerate))


def extract_audio_from_video(video_path: str, outpath: str = None) -> None:
    """extracts audio from a video
    video_path: path to video file in mp4
    outpath: path to save the audio file
    """

    if outpath is None:
        outpath = OUTPUT_DIR

    if not os.path.isdir(outpath):
        os.makedirs(outpath)

    # ffmpeg -i input.mp4 -vn -acodec copy output-audio.aac
    # command = "ffmpeg -i C:/test.mp4 -ab 160k -ac 2 -ar 44100 -vn audio.wav"
    # subprocess.call(command, shell=True)
    os.system(f"ffmpeg -i {video_path} -vn -acodec copy {outpath}/audio.wav")


def upscale_images_in_folder(image_folder: str, outpath: str = None):
    print(f"upscaling images in {image_folder}")
    path_to_real_esrgan = (
        MODELS_DIR + "/upscaling/realesrgan-ncnn-vulkan/realesrgan-ncnn-vulkan.exe"
    )

    # get imags
    lowres_imgs = get_files_in_dir(image_folder, [".jpeg", ".jpg", ".png"])
    # order images by creation date otherwise the video will be out of order
    lowres_imgs.sort(key=os.path.getmtime)

    # create output dir
    if outpath is None:
        outpath = os.path.join(image_folder, "upscaled")
        if not os.path.isdir(outpath):
            os.makedirs(outpath)

    # convert images with realesrgan
    for limg in tqdm(lowres_imgs):
        out_img = outpath + "/" + os.path.basename(limg)
        cmd = f"{path_to_real_esrgan} -i {limg} -o {out_img}"
        os.system(cmd)


def upscale_video(video_path: str, outpath: str = None):
    """
    Uses ESRGAN to upscale video. The audio is reaplied to the upscaled video.
    """
    # video2images(video_path, outpath)
    # extract_audio_from_video(video_path, outpath)
    # upscale images
    outupscaled = outpath + "/upscaled"
    # upscale_images_in_folder(outpath, outupscaled)

    # make video from images
    image_paths = get_files_in_dir(outupscaled, [".jpeg", ".jpg", ".png"])
    # get framerate from file if it exists
    if os.path.isfile(os.path.join(outpath, "_framerate.txt")):
        with open(os.path.join(outpath, "_framerate.txt"), "r") as f:
            frame_rate = int(float(f.readline()))
    else:
        print("Warning: Could not find framerate file. Using default framerate 60.")
        frame_rate = 60

    make_video_from_images(
        image_paths, outupscaled + "/upscaled.mp4", frame_rate=frame_rate
    )



import re
import os
import urllib
import unicodedata
import glob


def encode_path_safe(filename: str, allow_unicode=False):
    """
    Makes a string path safe by removing / replacing not by the os allowed patterns.
    This converts:
    spaces 2 dashes, repeated dashes 2 single dashes, remove non alphanumerics, underscores, or hyphen, string 2 lowercase
    strip
    """
    filename = str(filename)
    if allow_unicode:
        filename = unicodedata.normalize('NFKC', filename)
    else:
        filename = (
            unicodedata.normalize('NFKD', filename)
            .encode('ascii', 'ignore')
            .decode('ascii')
        )
    filename = re.sub(r'[^\w\s-]', '', filename.lower())
    return re.sub(r'[-\s]+', '-', filename).strip('-_')


def get_files_in_dir(path: str, extensions: list | str = None) -> list:
    """returns all files in a directory filtered by extension list"""
    if not os.path.isdir(path):
        print(f"{path} is not a directory. Returning empty list")
        return []

    files = []
    if extensions is None:
        # return all files
        files = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    else:
        # return files filtered by extensions
        if extensions is str:
            extensions = [extensions]
        for ext in extensions:
            files.extend(glob.glob(os.path.join(path, "*" + ext)))

    return files


def download_file(download_url: str, save_path: str):
    model_dir = os.path.dirname(save_path)
    if not os.path.isdir(model_dir):
        os.makedirs(model_dir, exist_ok=True)

    if not os.path.isfile(save_path):
        print(f'Downloading {download_url}')
        urllib.request.urlretrieve(download_url, save_path)
        print(f'Downloaded {download_url}')
    return save_path



from typing import Union, List
import os
import numpy as np
import insightface
from insightface.app.common import Face

from face2face.core.file_writable_face import FileWriteableFace
from face2face.utils import get_files_in_dir


def get_face_analyser(model_path: str, providers, det_size=(320, 320)):
    """Get the face analyser model. The face analyser detects faces and extracts facial features."""
    # load default face analyser if model_path is None
    face_analyser = insightface.app.FaceAnalysis(
        name="buffalo_l", root=f"{model_path}./checkpoints", providers=providers
    )
    face_analyser.prepare(ctx_id=0, det_size=det_size)
    # TODO: load face analyser from file if model_path is not None

    return face_analyser


def load_reference_face_from_file(face_embedding_file_path: str) -> Union[List[Face], None]:
    """
    Load a reference face from a file.
    :param face_embedding_file_path: the file path of the reference face
    :return: the reference face
    """
    if not os.path.exists(face_embedding_file_path):
        print(f"Reference face {face_embedding_file_path} not found.")
        return None

    try:
        # note: potential security issue, if the file was not created with face2face
        embedding = np.load(face_embedding_file_path, allow_pickle=True)
        if len(embedding) > 0:
            embedding = [FileWriteableFace.to_face(face) for face in embedding]

        return embedding
    except Exception as e:
        print(f"Error loading reference face {face_embedding_file_path}: {e}")


def load_reference_faces_from_folder(folder_path: str) -> dict:
    """
    Load reference faces from a folder. The folder must contain .npz files with the reference faces.
    The file name will be used as the reference face name.
    :param folder_path: the folder path
    :return:
    """
    files = get_files_in_dir(folder_path, [".npz"])
    reference_faces = {}
    for file in files:
        face_name = os.path.basename(file)[:-4]
        embedding = load_reference_face_from_file(file)
        if embedding is None:
            continue
        reference_faces[face_name] = embedding

    return reference_faces


# ordinary imports
from io import BytesIO
from typing import List, Union, Tuple
import copy
import os

# image processing imports
import numpy as np
# model imports
import insightface
import onnxruntime
from insightface.app.common import Face

from face2face.core.file_writable_face import FileWriteableFace
from face2face.core.f2f_loader import get_face_analyser, load_reference_face_from_file
from face2face.utils.utils import encode_path_safe, download_file
from face2face.settings import MODELS_DIR, REF_FACES_DIR, MODEL_DOWNLOAD_URL



class Face2Face:
    def __init__(self,
                 model_path: str = None,
                 inswapper_model_name: str = "inswapper_128.onnx",
                 reference_faces_folder: str = None,
                 ):
        """
        :param model_path: the folder where inswapper model and the buffalo_l model is stored.
        :param inswapper_model_name:
        """

        # prepare models
        if model_path is None:
            model_path = os.path.join(MODELS_DIR, 'insightface')

        # download models if not existing
        model_file_path = os.path.join(model_path, inswapper_model_name)
        if not os.path.isfile(model_file_path):
            download_file(MODEL_DOWNLOAD_URL, model_file_path)

        if reference_faces_folder is None:
            reference_faces_folder = REF_FACES_DIR

        self.providers = onnxruntime.get_available_providers()
        self._face_analyser = get_face_analyser(model_path, self.providers)
        self._face_swapper = insightface.model_zoo.get_model(model_file_path)

        # face swapper has the option to swap from image to image or
        # to have a reference images with reference faces and apply them to an image
        # they dict has structure {face_name: detected faces }
        self._reference_faces_folder = reference_faces_folder
        self.reference_faces = {}

    def get_many_faces(self, frame: np.ndarray) -> Union[List | None]:
        """
        get faces from left to right by order
        """
        try:
            face = self._face_analyser.get(frame)
            return sorted(face, key=lambda x: x.bbox[0])
        except IndexError:
            return None

    def _swap_detected_faces(self, source_faces, target_faces, target_image: np.array) -> np.array:
        """
        Changes the face(s) of the target image to the face(s) of the source image.
        if there are more target faces than source faces, the source face index is reset
        source_faces: the source faces
        target_image: the target image in BGR format (read with cv2)
        """
        if source_faces is None or len(source_faces) == 0:
            raise Exception("No source faces found!")

        if target_faces is None or len(target_faces) == 0:
            print(f"No face found in image. Return image as is")
            return target_image

        result = copy.deepcopy(target_image)

        # iter through all target faces and swap them with the source faces
        # if there are more target faces than source faces, the source face index is reset
        for target_index in range(len(target_faces)):
            source_index = target_index % len(source_faces)  # 6 % 5 = 1 and 1 % 5 = 1 ...
            result = self._face_swapper.get(
                result,  # in place operation
                target_faces[target_index],
                source_faces[source_index],
                paste_back=True,
            )

        return result

    def swap_one_image(
        self, source_image: np.array, target_image: np.array
    ) -> np.array:
        """
        Changes the face(s) of the target image to the face(s) of the source image.
        source_image: the source image in BGR format (read with cv2)
        target_image: the target image in BGR format (read with cv2)
        """

        # get the bounding box of the faces
        source_faces = self.get_many_faces(source_image)

        if source_faces is None:
            raise Exception("No source faces found!")

        target_faces = self.get_many_faces(target_image)
        return self._swap_detected_faces(source_faces, target_faces, target_image)

    def load_reference_embedding(self, face_name: str) -> Union[List[Face], None]:
        """
        Load a reference face embedding from a file.
        :param face_name: the name of the reference face embedding
        :return: the embedding of the reference face(s)
        """
        # check if is already in ram. If yes return that one
        embedding = self.reference_faces.get(face_name, None)
        if embedding is not None:
            return embedding

        # load from file
        file = os.path.join(self._reference_faces_folder, f"{face_name}.npz")
        embedding = load_reference_face_from_file(file)

        if embedding is None:
            raise ValueError(f"Reference face {face_name} not found. "
                             f"Please add the reference face first with add_reference_face")

        # add to memory dict
        self.reference_faces[face_name] = embedding
        return embedding

    def add_reference_face(self, face_name: str, ref_image: np.array, save=False) -> Tuple[str, np.array]:
        """
        Add a reference face to the face swapper. The face swapper will use this face to swap it to other images.
        Use the method swap_from_reference_face to swap the reference face to other images.
        :param face_name: how the reference face is called
        :param ref_image: the image to get the faces from
        :param save: if True, the reference face will be saved to the reference_faces folder and available next startup
        :return: the savely encoded face name and the reference face
        """
        face_name = encode_path_safe(face_name)

        self.reference_faces[face_name] = self.get_many_faces(ref_image)

        # make classes pickle safe
        for i, face in enumerate(self.reference_faces[face_name]):
            self.reference_faces[face_name][i] = FileWriteableFace(face)

        # save face to virtual file
        virtual_file = BytesIO()
        np.save(virtual_file, self.reference_faces[face_name], allow_pickle=True)
        virtual_file.seek(0)
        if save:
            if not os.path.isdir(REF_FACES_DIR):
                os.makedirs(REF_FACES_DIR)

            filename = os.path.join(REF_FACES_DIR, f"{face_name}.npz")
            if os.path.isfile(filename):
                print(f"Reference face {face_name} already exists. Overwriting.")

            with open(filename, "wb") as f:
                f.write(virtual_file.getbuffer())

        virtual_file.seek(0)
        return face_name, virtual_file

    def swap_from_reference_face(self, face_name: str, target_image: Union[np.array, list]) -> np.array:
        """
        Changes the face(s) of the target image to the face of the reference image.
        :param face_name: the name of the reference face
        :param target_image: the target image in BGR format (read with cv2). Can be a list of images
        :return: the swapped image
        """
        face_name = encode_path_safe(face_name)

        # if target_image is a list of images, swap all images
        if isinstance(target_image, list):
            return list(self.swap_generator(face_name, target_image))

        # swap single image
        source_faces = self.load_reference_embedding(face_name)
        target_faces = self.get_many_faces(target_image)
        return self._swap_detected_faces(source_faces, target_faces, target_image)

    def swap_generator(self, face_name: str, target_generator):
        """
        Changes the face(s) of each image in the target_img_generator to the face of the reference image.
        :param face_name: the name of the reference face
        :param target_generator: a generator that yields images in BGR format (read with cv2).
            Or a video_stream that yields (image, audio) like in media_toolkit.
        :return: a generator that yields the swapped images or a tuple (image, audio)
        """
        face_name = encode_path_safe(face_name)
        source_faces = self.load_reference_embedding(face_name)

        for i, target_image in enumerate(target_generator):
            # check if generator yields tuples (video, audio) or only images
            audio = None
            if isinstance(target_image, tuple) and len(target_image) == 2:
                target_image, audio = target_image

            try:
                target_faces = self.get_many_faces(target_image)
                swapped = self._swap_detected_faces(source_faces, target_faces, target_image)
                if audio is not None:
                    yield swapped, audio

                yield swapped
            except Exception as e:
                print(f"Error in swapping frame {i} to {face_name}: {e}. Skipping image")
                if audio is not None:
                    yield target_image, audio

                yield np.array(target_image)

    def swap_video(self, face_name: str, target_video, include_audio: bool = True):
        """
        Swaps the face of the target video to the face of the reference image.
        :param face_name: the name of the reference face embedding
        :param target_video: the target video. Path to the file or VideoFile object
        :param include_audio: if True, the audio will be included in the output video
        """
        try:
            from media_toolkit import VideoFile
        except:
            raise ImportError("Please install socaity media_toolkit to use this function")

        if isinstance(target_video, str):
            target_video = VideoFile().from_file(target_video)

        if not isinstance(target_video, VideoFile):
            raise ValueError("target_video must be a path or a VideoFile object")

        gen = target_video.to_video_stream() if include_audio else target_video.to_image_stream()

        new_video = VideoFile().from_video_stream(
            video_audio_stream=self.swap_generator(face_name, gen),
            frame_rate=target_video.frame_rate,
            audio_sample_rate=target_video.audio_sample_rate
        )
        return new_video






import os

ROOT_DIR = os.getenv("ROOT_DIR", os.path.dirname(os.path.abspath(__file__)))
MODELS_DIR = os.path.join(ROOT_DIR, "models")
OUTPUT_DIR = os.path.join(ROOT_DIR, "output")
REF_FACES_DIR = os.getenv("REF_FACES_DIR", os.path.join(ROOT_DIR, "face_embeddings"))
MODEL_DOWNLOAD_URL = "https://github.com/facefusion/facefusion-assets/releases/download/models/inswapper_128.onnx"

PORT = 8020
PROVIDER = os.getenv("PROVIDER", "fastapi")
